{
MIDI files decoding and playing code for Win32
----------------------------------------------

Based on MSVC++ source code of TMIDI by Tom Grandgent

Rewrited on Object Pascal for Delphi 7 by Sergey Bulba

(c)2006 S.V.Bulba
http://bulba.at.kz/
vorobey@mail.khstu.ru

This is second edition of rewriting and removing extra code
from original TMIDI. Only playing, seeking and visualisation
code is used. Must play all MIDI files of types from 0 to 2
(.MID and .RMI). Uses the lowest level of Win32 Multimedia
System. Some bugs of TMIDI was fixed here also.
}

unit Midi;

interface

uses Windows,Messages,MMSystem,Sysutils,StdCtrls;

const
 MThd = $6468544D;
 MTrk = $6B72544D;
 RIFF = $46464952;
 MIDI_STANDARD_NONE=1;
 MIDI_STANDARD_GM=2;
 MIDI_STANDARD_GS=4;
 MIDI_STANDARD_XG=8;
 MIDI_STANDARD_MT32=16;
 MAX_MIDI_WAIT=1000.0;	// Max # of milliseconds to sleep in main playback loop
 MAX_MIDI_TRACKS=512;		// Max # of tracks Midi.pas will read from a MIDI file
 MAX_PITCH_BEND=$2000;	// Max value for pitch bends (this is a MIDI constant, do not change!)

type
 PByteArray = ^TByteArray;
 TByteArray = packed array[0..65535] of byte;
 Pmidi_device_t = ^midi_device_t;
 midi_device_t = record
	outcaps:MIDIOUTCAPS;				// Output device caps
	next:Pmidi_device_t;
 end;

 channel_state_t = record
	// State values
  changed_pitch_bend:integer;
	normal_program:integer;					// Normal program used by this channel
	notes:packed array[0..127] of byte;			// Current velocity for each possible note (0 is off)
	note_count:integer;						// Number of notes currently on
	// Historical values
	last_bank:shortint;				// Last bank set from this track
	last_pitch_bend:integer;			// Last pitch bend set from this track
 end;

 midi_state_t = record
	analyzing:boolean;						// Is this song currently being analyzed?
	seek_sliding:boolean;					// Is the seek bar sliding? (user holding down button)
	seeking:boolean;						// Is this song currently seeking?
	seek_to:double;						// Place to seek to (in milliseconds)
	stop_requested:boolean;					// Has there been a request to stop the playback thread?
	playing:boolean;						// Is the playback thread playing?
	paused:boolean;							// Is the playback thread paused?
	finished_naturally:boolean;				// Did the current MIDI file finish naturally?
	loop_count:integer;						// Number of times left to loop the playback
	starttime:double;					// Starting playback time
	curtime:double;						// Current playback time
  elapsedtime:longword;
	tempo:longword;					// Current tempo
	tick_length:double;					// Current tick length in milliseconds
	song_length:integer;					// Length of song in seconds
	channels:array[0..15] of channel_state_t;		// Channel state structs
	midi_standard:integer;					// MIDI standard used by this MIDI file (as sysex indicates)
	filename:string;
 end;

 midi_header_t = packed record
	id:longword;							// MIDI header ID ("MThd")
	header_size:longword;			// Size of header in bytes
	file_format:word;			// MIDI file type (0, 1, or 2)
	num_tracks:word;			// Number of tracks in MIDI file
	num_ticks:word;			// Number of ticks per quarter note (used to calc tempo)
 end;

 track_header_t = record
	length:longword;				// Length of track chunk in bytes
	data:PByteArray;				// Pointer to beginning of track chunk
	dataptr:PByteArray;				// Pointer to current location in track chunk
	trigger:double;						// Ending time of current event
	dt:double;							// Length of current event in ticks
	enabled:boolean;						// Is this track enabled?
	tracknum:word;			// Number of this track
	lastcmd:byte;				// Last command byte (used in running mode)
 end;

 Pmidi_text_t = ^midi_text_t;
 midi_text_t = record
	midi_text:string;					// MIDI text string
	text_type:integer;						// Type of text event
	next:Pmidi_text_t;
 end;

procedure MIDIEnumDevices(cb:TComboBox);
procedure init_midi_out(d:integer);
procedure close_midi_out;
function load_midi(filename:string):boolean;
procedure all_notes_off_channel(channel:integer);
procedure all_notes_off;
function process_midi_event(var th:track_header_t):boolean;
procedure set_tempo(new_tempo:integer);
procedure note_on(on_:boolean;note,velocity,chan:byte);
procedure set_channel_program(channel,prog:byte);
procedure set_channel_pitch(channel:byte;intd:integer);
procedure output_sysex_data(init:byte; data:PByteArray; length:integer);
procedure update_note_volume(channel, note, volume:byte);
function MIDIThreadFunc(a:pointer):dword;stdcall;
procedure StopMIDIThread;

var
 midi_devices:Pmidi_device_t = nil;		// MIDI devices
 hout:HMIDIOUT = 0;
 ms:midi_state_t;						// MIDI state struct
 mh:midi_header_t;						// MIDI file header struct
 th:array[0..MAX_MIDI_TRACKS-1] of track_header_t;		// Track state structs
 midi_text_events:Pmidi_text_t = nil;	// MIDI text events

// Timing variable
PeriodMin:longword = 1;

MIDIThreadID:DWORD;
MIDIThreadH:THandle = 0;

MIDILoop:boolean = False;

hwndApp:THandle;

implementation

var
	infile:file;
  last_tempo:integer;
  sysex_data_buf:array of byte;

function IntelWord(Wrd:word):word;
asm
xchg al,ah
end;

function IntelDWord(DWrd:dword):dword;
asm
xchg al,ah
ror eax,16
xchg al,ah
end;

function MAKEBYTE(a,b:byte):byte;
begin
Result := (a and 15) or (b shl 4)
end;

procedure MIDIEnumDevices(cb:TComboBox);
var
 outdevs:integer;
 outcaps:MIDIOUTCAPS;
 i,ret:integer;
 dev,ld:Pmidi_device_t;
begin
 ld := nil;
 cb.Items.Add('MIDI Mapper');
 outdevs := midiOutGetNumDevs;
 for i := 0 to outdevs - 1 do
	begin
		ret := midiOutGetDevCaps(i, @outcaps, sizeof(outcaps));
		if (ret <> MMSYSERR_NOERROR) then	continue;

		// Set up a device node for this device
    New(dev);
		Move(outcaps, dev.outcaps, sizeof(MIDIOUTCAPS));
		dev.next := nil;
		if ld = nil then
			midi_devices := dev
		else
			ld.next := dev;
		ld := dev;
    cb.Items.Add(outcaps.szPname)
  end;
cb.ItemIndex := 0
end;

procedure init_midi_out(d:integer);
begin

	if hout <> 0 then
	 begin
		midiOutClose(hout);
		hout := 0;
	 end;

	if midiOutOpen(@hout, d, 0, 0, 0) <> MMSYSERR_NOERROR then
	 begin
		hout := 0;
		MessageBox(hwndApp,'Unable to open MIDI-out device:', 'midiOutOpen failed...', MB_ICONERROR);
   end;
end;

procedure close_midi_out;
var
 i:integer;
begin

	if hout <> 0 then
   begin
		all_notes_off;
		midiOutReset(hout);
    i := 0;
		while (midiOutClose(hout) <> MMSYSERR_NOERROR) and (i < 10) do
     begin
      inc(i);
			Sleep(200);
     end;
		if i = 10 then
			MessageBox(hwndApp, 'Unable to close MIDI-out device!', 'DMIDI Error', MB_ICONERROR);
		hout := 0;
   end;
end;

// Turns all notes off on a specific channel
procedure all_notes_off_channel(channel:integer);
begin
	// Clear saved note velocities for this channel
  FillChar(ms.channels[channel].notes,128,0);
	ms.channels[channel].note_count := 0;
	// Turn off notes on the MIDI-out device
	if hout <> 0 then
	 begin
		// Better safe than sorry!
    midiOutShortMsg(hout, MAKELONG(MAKEWORD($B0 + channel, 120), MAKEWORD(0, 0)));
//    midiOutShortMsg(hout, MAKELONG(MAKEWORD($B0 + channel, 121), MAKEWORD(0, 0)));
//    midiOutShortMsg(hout, MAKELONG(MAKEWORD($B0 + channel, 123), MAKEWORD(0, 0)));
	 end;
end;

// Turns all notes off on all channels
procedure all_notes_off;
var
 c:integer;
begin
	for c := 0 to 15 do
   all_notes_off_channel(c);
end;

// Adds a new node to the list of MIDI text events
function new_midi_text(midi_text:string; text_type:integer):Pmidi_text_t;
var
	p,n:Pmidi_text_t;
	ch:integer;
begin

	// Allocate memory for the new node
  new(n);

	// If the list is empty.. add this as the top node
	if midi_text_events = nil then
		midi_text_events := n
	else
	 begin
		// Otherwise find the end of the list and add this node to the end
    p := midi_text_events; while p.next <> nil do p := p.next;
		p.next := n;
	 end;

	// Initialize data for this node
	n.text_type := text_type;
	n.next := nil;

	if midi_text <> '' then
	 begin
		// Sneaky trick to detect MT-32 MIDI files!
		if (ms.midi_standard = MIDI_STANDARD_NONE) and (pos('MT-32',midi_text) <> 0) then
			ms.midi_standard := MIDI_STANDARD_MT32;

		// Cool trick to break up lines with CR/LF pairs
		// Search for a CR or LF
		ch := Pos(midi_text, #13);
		if ch = 0 then ch := Pos(midi_text, #10);
		// Found one?  Good.. let's break up the line into multiple text entries!
		if ch <> 0 then
		 begin
			// Terminate the string at this character and move to the next character
			midi_text[ch] := #0; inc(ch);
			// See if there's another CR/LF to seek past
			if midi_text[ch] in [#13,#10] then inc(ch);
			// Make a new text entry with what we just seeked past
			new_midi_text(PChar(@midi_text[ch]), text_type);
		 end;

		n.midi_text := PChar(midi_text)
	 end;

	// Return a pointer to the new node
	Result := n;
end;

// Removes all nodes in the list of MIDI text events
procedure kill_all_midi_text;
var
 p,next:Pmidi_text_t;
begin
	p := midi_text_events;
	while p <> nil do
   begin
		next := p.next;
		Dispose(p);
		p := next;
	 end;
	midi_text_events := nil;
end;

procedure check_midi_standard(data:PByteArray;len:integer);
begin
if len = 0 then exit;
	case data[0] of								// Manufacturer ID
		$41:	// Roland
     begin
      if len < 3 then exit;
			case data[2] of
				$16:	// MT-32
					ms.midi_standard := MIDI_STANDARD_MT32;
				$42:	// GS Message
					ms.midi_standard := MIDI_STANDARD_GS;
      end;
     end;
		$43:	// Yamaha
     begin
      if len < 3 then exit;
			case data[2] of
				$4C:	// XG Message
					ms.midi_standard := MIDI_STANDARD_XG;
			end;
     end; 
		$7E,	// General MIDI
		$7F:
			ms.midi_standard := MIDI_STANDARD_GM;
  end;
end;

function read_short:word;
var
 v:word;
begin
	BlockRead(infile,v,2);
	Result := IntelWord(v)
end;

function read_int:longword;
var
 v:longword;
begin
	BlockRead(infile,v,4);
	Result := IntelDWord(v)
end;

function read_bytes_mem(var th:track_header_t; buf:PChar; num:integer):integer;
begin
  if longword(th.dataptr) - longword(th.data) + longword(num) > th.length then
   begin
  	MessageBox(hwndApp, PChar('File damaged: ' + ms.filename), 'read_bytes_mem Error', MB_ICONERROR);
    halt(1)
	 end;
  Move(th.dataptr^,buf^,num);
	inc(integer(th.dataptr),num);
	Result := num
end;

function read_byte_mem(var th:track_header_t):byte;
begin
  if longword(th.dataptr) - longword(th.data) >= th.length then
   begin
  	MessageBox(hwndApp, PChar('File damaged: ' + ms.filename), 'read_byte_mem Error', MB_ICONERROR);
    halt(1)
	 end;
  Result := th.dataptr^[0];
  inc(integer(th.dataptr));
end;

function read_int_mem(var th:track_header_t):longword;
begin
  if longword(th.dataptr) - longword(th.data) + 4 > th.length then
   begin
  	MessageBox(hwndApp, PChar('File damaged: ' + ms.filename), 'read_int_mem Error', MB_ICONERROR);
    halt(1)
	 end;
	Result := IntelDWord(PLongWord(th.dataptr)^);
  inc(integer(th.dataptr),4);
end;

function read_short_mem(var th:track_header_t):word;
begin
  if longword(th.dataptr) - longword(th.data) + 2 > th.length then
   begin
  	MessageBox(hwndApp, PChar('File damaged: ' + ms.filename), 'read_short_mem Error', MB_ICONERROR);
    halt(1)
	 end;
	Result := IntelWord(PWord(th.dataptr)^);
  inc(integer(th.dataptr),2);
end;

function read_vlq_mem(var th:track_header_t):longword;
var
 value:longword;
 c:byte;
 p:PByteArray;
 i:integer;
begin
	p := th.dataptr;

  if longword(p) - longword(th.data) >= th.length then
   begin
  	MessageBox(hwndApp, PChar('File damaged: ' + ms.filename), 'read_vlq_mem Error', MB_ICONERROR);
    halt(1)
	 end;

        i := 3;
    value := p^[0];
    inc(integer(p));
    if (value and $80) <> 0 then
     begin
        value := value and $7f;
        repeat

  if longword(p) - longword(th.data) >= th.length then
   begin
  	MessageBox(hwndApp, PChar('File damaged: ' + ms.filename), 'read_vlq_mem Error', MB_ICONERROR);
    halt(1)
	 end;

         c := p^[0];
         inc(integer(p));
    		 value := (value shl 7) + (c and $7f);
         dec(i);
        until (c and $80 = 0)// or (i = 0)
     end;
	th.dataptr := p;
  if i < 0 then //Error !!!
   begin
  	MessageBox(hwndApp, PChar('File damaged: ' + ms.filename), 'read_vlq_mem > 4 bytes Error', MB_ICONERROR);
    halt(1);
    value := value and $7F;
	 end;

  Result := value;
end;

function analyze_midi:boolean;
var
 tracks,i{,num_events}:integer;
 curtime,starttime,nexttrigger:double;
 first_pass:boolean;
begin
	tracks := mh.num_tracks;
//	num_events := 0;

	// Initialize playback parameters
	ms.midi_standard := MIDI_STANDARD_NONE;
	ms.stop_requested := False;
	ms.tempo := 500000 div 1000;
	ms.tick_length := ms.tempo / mh.num_ticks;
	ms.analyzing := True;

	// Initialize track parameters
	for i := 0 to tracks - 1 do
	 begin
		th[i].lastcmd := 0;
		th[i].enabled := True;
		th[i].trigger := 0.0;
		th[i].dataptr := th[i].data;
		th[i].tracknum := i;
	 end;

	// Initialize historical channel values
	for i := 0 to 15 do
 	 ms.channels[i].last_bank := -1;

	// Record starting time
  starttime := 0.0;
	curtime := 0.0;
	ms.starttime := starttime;

	// Loop until we've been asked to stop
	first_pass := True;
	while not ms.stop_requested do
	 begin
		// Schedule the next "wakeup time" MAX_MIDI_WAIT milliseconds into the future
		nexttrigger := curtime + MAX_MIDI_WAIT;
		// See if any tracks have pending data
		ms.stop_requested := True;
		for i := 0 to tracks - 1 do
     begin
			// Continue on to the next track if this track is disabled
			if not th[i].enabled then
				continue
			else
				ms.stop_requested := False;

			// Read the next event's delta time
			if (first_pass) then
			 begin
				th[i].dt := read_vlq_mem(th[i]) * ms.tick_length;
				th[i].trigger := curtime + th[i].dt;
			 end;

			// Process MIDI events until one is scheduled for a time in the future
			while (curtime >= th[i].trigger) do
			 begin

				// Process the event for this track
				ms.curtime := curtime;
				if (process_midi_event(th[i])) then
         begin
          Result := True;
					exit;
         end;
//				inc(num_events);

				// Check for end of track
				if not th[i].enabled then	break;

				if longword(th[i].dataptr) >= (longword(th[i].data) + th[i].length) then
				 begin
					th[i].enabled := false;
					break;
				 end;

				// Read the next event's delta time
				th[i].dt := read_vlq_mem(th[i]) * ms.tick_length;
				th[i].trigger := th[i].trigger + th[i].dt;
			 end;
			// Check for end of track
			if not th[i].enabled then	continue;

			// See if this track's trigger is the more recent than the next scheduled trigger
			// If so, make it the next trigger
			if (th[i].trigger < nexttrigger) then
				nexttrigger := th[i].trigger;
		 end;
		first_pass := False;
		curtime := nexttrigger;
	 end;
	// Analysis complete

	// Store song length and # of events
	ms.song_length := round(curtime / 1000.0);

	Result := False;
end;

function load_midi(filename:string):boolean;
var
	id:longword;
	s:string;
  buf:array[0..1023] of byte;
  track,i,len:longword;
  success:boolean;

begin
  Result := False;

	track := 0;
	success := False;

	// Save the given filename
  ms.filename := filename;

	// Dump any previously-loaded MIDI text and sysex events
	kill_all_midi_text;

	// Unload any previously-loaded MIDI data
	if mh.num_tracks <> 0 then
		for i := 0 to mh.num_tracks - 1 do
		 begin
			th[i].dataptr := nil;
			if th[i].data <> nil then FreeMem(th[i].data);
      FillChar(th[i],sizeof(track_header_t),0);
		 end;

  last_tempo := 0;

  FileMode := 0;
	AssignFile(infile,filename);
  Reset(infile,1);

	try

	// Read the MIDI file header
	BlockRead(infile, id, 4);

	// Check to see if the "MThd" ID is correct
	if MThd <> id then
	 begin
		// No?  Check for a RIFF header (RMID format)
		if RIFF = id then
		 begin
			// Yes!  Ok, skip past 4 byte filesize + 4 byte RMID
			BlockRead(infile, id, 4);
			BlockRead(infile, id, 4);
			BlockRead(infile, id, 4);
			BlockRead(infile, id, 4);
			// Now read the MIDI file header
			BlockRead(infile, id, 4);
			// Now look for MThd
			if MThd <> id then
			 begin
        s := filename + ' is not a MIDI file because it does not begin with "MThd".';
				MessageBox(hwndApp, PChar(s), 'DMIDI Error', MB_ICONERROR);
        exit;
			 end
     end
		else
		 begin
			// STILL no MThd?  Search through the beginning of the file for MThd
			Seek(infile,0);
			BlockRead(infile,buf,1024,len);
      i := 0;
			if len > 4 then
       begin
				while i < len - 4 do
         begin
					if MThd = PLongWord(buf[i])^ then break;
          inc(i)
         end;
       end;
			if (len <= 4) or (i = len - 4) then
			 begin
        s := filename + ' is not a MIDI file because it does not begin with "MThd".';
				MessageBox(hwndApp, PChar(s), 'DMIDI Error', MB_ICONERROR);
        exit;
			 end
			else
			 begin
				Seek(infile, i);
				BlockRead(infile,id,4);
			 end
		 end
	 end;

	// Read the header size
	mh.header_size := read_int;

	// Identify the file format
	mh.file_format := read_short;
  if not mh.file_format in [0..2] then
   begin
		MessageBox(hwndApp,PChar('Unknown MIDI file type: '
                      + IntToStr(mh.file_format)), 'DMIDI Error', MB_ICONERROR);
		exit
   end;

	// Output other information present in the header
	mh.num_tracks := read_short;
	mh.num_ticks := read_short;

	// Read tracks
	while not eof(infile) do
	 begin
		// Read the track header
		BlockRead(infile, id, 4);
		// Check to see if the "MTrk" ID is correct
		if MTrk <> id then
		 begin
      s := 'Track header ' + IntToStr(track + 1) + ' is not correct because it does not begin with "MTrk".';
			if track < mh.num_tracks then	MessageBox(hwndApp, PChar(s), 'DMIDI Warning', MB_ICONWARNING);
			break;
     end;
		// Read track length in bytes
		th[track].length := read_int;

		// Read the track data into memory
    GetMem(th[track].data,th[track].length);
		if th[track].data = nil then
		 begin
			MessageBox(hwndApp, 'GetMem failed trying to allocate memory for track data', 'DMIDI Error', MB_ICONERROR);
			break;
		 end;
		BlockRead(infile,th[track].data^,th[track].length);
		inc(track);
		if (track >= MAX_MIDI_TRACKS) then
		 begin
			MessageBox(hwndApp, PChar('Warning: Max number of MIDI tracks reached (' +
           IntToStr(MAX_MIDI_TRACKS) +
        ').'#13#10'I refuse to read any more from this file.'), 'DMIDI Warning', MB_ICONWARNING);
			break;
		 end;
	 end;
	success := True;

	finally

		mh.num_tracks := track;
		CloseFile(infile);
		if (success) and not analyze_midi then
     Result := True
		else
     MessageBox(hwndApp,PChar(filename + ' file load error'),'Error',MB_ICONERROR);
	end;

end;

function process_midi_event(var th:track_header_t):boolean;
var
 text:PChar;
 id:longword;
 i:integer;
 len:longword;
 cmd, d1, d2:byte;
 intd:longword;
 channel:integer;
 sysexptr:PByteArray; // Pointer to beginning of sysex data
begin

	// Read the MIDI event command
	cmd := read_byte_mem(th);

	// Handle running mode
	if (cmd < 128) then
	 begin
		dec(integer(th.dataptr));
		cmd := th.lastcmd;
    if cmd < 128 then
		 begin
			MessageBox(hwndApp, 'Broken File (cmd < 128)', 'MIDI Read Error', MB_ICONERROR);
			Result := True;
      exit
		 end;
	 end
	else if cmd <> 255 then
   th.lastcmd := cmd;

	if (cmd = $FF) then		// Meta-event
	 begin
		cmd := read_byte_mem(th);
		len := read_vlq_mem(th);
		if len > th.length - (longword(th.dataptr) - longword(th.data)) then
		 begin
			MessageBox(hwndApp, 'Meta-event length error', 'MIDI Read Error', MB_ICONERROR);
			Result := True;
      exit
		 end;
		case (cmd) of

			$01,		// Text
			$02,		// Copyright info
			$03,		// Sequence or track name
			$04,		// Track instrument name
			$05,		// Lyric
			$06,		// Marker
			$07:		// Cue point
       begin
				// Read the text
				GetMem(text,len + 1);
				read_bytes_mem(th, text, len);
				text[len] := #0;
				// Save the text in the list of text events
				if (ms.analyzing and not ms.seeking) then
					new_midi_text(text, cmd);
        FreeMem(text);
  		 end;
			$2F:		// End of track
       begin
				th.enabled := False;
        for i := 0 to len - 1 do
  			  read_byte_mem(th);
       end;
			$51:		// Set tempo
       begin
        if Len >= 3 then
         begin
   				read_bytes_mem(th, @id, 3);
          intd := IntelDWord(id shl 8);
		  		set_tempo(intd);
          for i := 0 to len - 4 do
           read_byte_mem(th);
         end
        else
         for i := 0 to len - 1 do
  			  read_byte_mem(th);
			 end;
 			else		// Unknown or not need for me
       for i := 0 to len - 1 do
 				read_byte_mem(th);
		end;
	 end
	else
	 begin
		channel := cmd and 15;
		// *** cool symmetric part follows: ***
		case cmd shr 4 of	// Normal event

			$08: // Note off
       begin
				d1 := read_byte_mem(th);
				d2 := read_byte_mem(th);
				if not ms.analyzing then
         note_on(FALSE, d1, d2, channel);
			 end;
			$09: // Note on
       begin
				d1 := read_byte_mem(th);
				d2 := read_byte_mem(th);
				if not ms.analyzing then
         note_on(TRUE, d1, d2, channel);
			 end;
			$0A: // Key after-touch
       begin
				d1 := read_byte_mem(th);
				d2 := read_byte_mem(th);
        if not ms.analyzing and (hout <> 0) then
         midiOutShortMsg(hout, MAKELONG(MAKEWORD(cmd, d1), MAKEWORD(d2, 0)));
			 end;
			$0B: // Control Change
       begin
				d1 := read_byte_mem(th);
				d2 := read_byte_mem(th);
        if not ms.analyzing or ms.seeking then
         begin
        	if d1 = 0 then
        	 ms.channels[channel].last_bank := d2;
          if hout <> 0 then
           for i := 0 to 1 do
            begin
             if midiOutShortMsg(hout, MAKELONG(MAKEWORD($B0 + channel, d1), MAKEWORD(d2, 0))) <> MIDIERR_NOTREADY	then break;
             sleep(10)
            end;
         end;
			 end;
			$0C: // Program Change
       set_channel_program(channel, read_byte_mem(th));
			$0D: // Channel after-touch
       begin
				d1 := read_byte_mem(th);
				if not ms.analyzing then
				 midiOutShortMsg(hout, MAKELONG(MAKEWORD(cmd, d1), MAKEWORD(0, 0)));
       end;
			$0E: // Pitch wheel
       begin
				intd := read_short_mem(th);
        set_channel_pitch(channel,intd);
			 end;
			$0F: // System message
       begin
				case channel of
          $01: //MTC Quarter Frame Message
 					 read_byte_mem(th);
					$02: //Song position
           begin
						read_byte_mem(th);
						read_byte_mem(th);
           end;
					$03: //Song select
           read_byte_mem(th);
					$00,
					$07:	// SYSEX data
           begin
						len := read_vlq_mem(th);
						if len > th.length - (longword(th.dataptr) - longword(th.data)) then
						 begin
							MessageBox(hwndApp, 'Sysex len error', 'DMIDI Read Error', MB_ICONERROR);
							Result := True;
              exit;
						 end;
						sysexptr := th.dataptr;
						if ms.analyzing and not ms.seeking then
             check_midi_standard(sysexptr, len);
						inc(integer(th.dataptr),len);
						if not ms.analyzing or ms.seeking then
             // Don't send sysex when seeking on the MT-32
             if not ((ms.midi_standard = MIDI_STANDARD_MT32) and ms.seeking) then
						 	output_sysex_data(channel and 1 xor 1,sysexptr,len);
					 end;
				end;
			 end;
	  end;
	 end;

	Result := False;
end;

procedure set_tempo(new_tempo:integer);
var
 old_tick_length:double;
 i:integer;
begin

	// Calculate new tick length
	old_tick_length := ms.tick_length;
	ms.tick_length := ((new_tempo / 1000) / mh.num_ticks);

	if (new_tempo = last_tempo) then exit;

	last_tempo := new_tempo;

	// Correct outstanding track triggers on each track
	for i := 0 to mh.num_tracks - 1 do
	 begin
		if not th[i].enabled then	continue;
		th[i].dt := (th[i].trigger - ms.curtime) / old_tick_length * ms.tick_length;
		th[i].trigger := ms.curtime + th[i].dt;
	 end;
end;

procedure note_on(on_:boolean;note,velocity,chan:byte);
var
 dwParam1:DWORD;
 channel:byte;
begin
	channel := chan;

  if velocity > 128 then
   MessageBox(hwndApp,PChar('velocity range error: ' + IntToStr(velocity)),'Error-warning',MB_ICONERROR);

	// Update display/historic values
  if on_ then
   update_note_volume(channel, note, velocity and 127)
  else
   update_note_volume(channel, note, 0);

	// If the bank for this channel is 127, treat it as a percussive channel! (XG)
	// This needs to be an option, in case someone actually has a real XG synth.
	if (ms.midi_standard = MIDI_STANDARD_XG) and (ms.channels[channel].last_bank = 127) then
		channel := 9;

  if on_ then
  	dwParam1 := MAKELONG(MAKEWORD(MAKEBYTE(channel, 9), note), MAKEWORD(velocity, 0))
  else
  	dwParam1 := MAKELONG(MAKEWORD(MAKEBYTE(channel, 8), note), MAKEWORD(velocity, 0));
	if hout <> 0 then	midiOutShortMsg(hout, dwParam1);
end;

procedure set_channel_program(channel,prog:byte);
begin
  if prog >= 128 then
   MessageBox(hwndApp,PChar('set_channel_program range error: ' + IntToStr(prog)),'Error-warning',MB_ICONERROR);

  ms.channels[channel].normal_program := prog;

	if not ms.analyzing and (hout <> 0) then
		midiOutShortMsg(hout, MAKELONG(MAKEWORD(MAKEBYTE(channel, $0C), prog), MAKEWORD(0, 0)));
end;

procedure set_channel_pitch(channel:byte;intd:integer);
var
 d1,d2:byte;
begin
ms.channels[channel].changed_pitch_bend := intd;
if not ms.analyzing then
 begin
  d2 := LOBYTE(intd);
  d1 := HIBYTE(intd);
	ms.channels[channel].last_pitch_bend := integer(integer(d2) shl 7 or d1) - MAX_PITCH_BEND;
  if (hout <> 0) then
   midiOutShortMsg(hout, MAKELONG(MAKEWORD($E0 + channel, d1), MAKEWORD(d2, 0)));
 end
end;

procedure output_sysex_data(init:byte; data:PByteArray; length:integer);
var
 mh:MIDIHDR;
 i:integer;
begin

	if hout = 0 then exit;

  inc(length,init);

	// Realloc the sysex buffer if it's too small
	if (length + 1 > System.Length(sysex_data_buf)) then
   SetLength(sysex_data_buf,length + 31);


  if init <> 0 then
	 sysex_data_buf[0] := $F0;					// Sysex begin command
	Move(data^, sysex_data_buf[init], length - init);	// Copy the sysex data into the local sysex buffer

	// Prepare the MIDI out header
	FillChar(mh, sizeof(mh), 0);
	mh.lpData := @sysex_data_buf[0];
	mh.dwBufferLength := length;
	mh.dwBytesRecorded := length;
	// Prepare the sysex buffer for output
	midiOutPrepareHeader(hout, @mh, sizeof(mh));

  for i := 0 to 1 do
   begin
  	// Send the sysex buffer!
  	if midiOutLongMsg(hout, @mh, sizeof(mh)) <> MIDIERR_NOTREADY then break;
    sleep(10)
   end;

	// Unprepare the sysex buffer
	midiOutUnprepareHeader(hout, @mh, sizeof(mh));
end;

// Updates the volume for a given note on a given channel
procedure update_note_volume(channel, note, volume:byte);
begin
	if volume <> 0 then
	 begin
		// Note on
		if ms.channels[channel].notes[note] = 0 then
			inc(ms.channels[channel].note_count);
	 end
	else
	 begin
		// Note off
		if ms.channels[channel].notes[note] <> 0 then
			dec(ms.channels[channel].note_count);
	 end;
	// Store the value
	ms.channels[channel].notes[note] := volume;
end;

function MIDIThreadFunc(a:pointer):dword;stdcall;
var
 tracks:integer;
 i:integer;
 curtime, starttime, pausetime, tmptime, timediff:double;
 nexttrigger:double;
 first_pass:boolean;
 tracks_active:boolean;
 seeking:boolean;
label BeginPlayback;
begin
  Result := STILL_ACTIVE - 1;
	tracks := mh.num_tracks;
	seeking := False;

	// Initialize seeking states
	ms.seeking := False;
	ms.seek_sliding := False;
	ms.seek_to := 0.0;

	// Init MIDI-out device
	init_midi_out(integer(a));
	if hout = 0 then exit;

	// Set playback thread priority
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST);

	// Initialize loop count
	ms.loop_count := 1;

  //Set max MMTimers precision
  timeBeginPeriod(PeriodMin);

	while ms.loop_count > 0 do
	 begin
    dec(ms.loop_count);
BeginPlayback:
		// Initialize tempo
		set_tempo(trunc(60000000.0 / ms.tempo));

		// Initialize playback parameters
		ms.analyzing := ms.seeking;
		ms.playing := True;
		ms.paused := False;
		ms.finished_naturally := FALSE;
		ms.stop_requested := False;
		ms.tempo := trunc(500000 / 1000);
		ms.tick_length := ms.tempo / mh.num_ticks;

		// Reset channel states
		for i := 0 to 15 do
		 begin
			// Reset # of notes currently being played on this channel
			ms.channels[i].note_count := 0;

      ms.channels[i].changed_pitch_bend := -1;
			// Reset program values
			ms.channels[i].normal_program := -1;
			// Reset note values
      FillChar(ms.channels[i].notes,sizeof(ms.channels[i].notes),0);
			// Reset historical values
			ms.channels[i].last_pitch_bend := 0;
		 end;

		// Initialize track parameters
		for i := 0 to tracks - 1 do
		 begin
			th[i].lastcmd := 0;
			th[i].enabled := True;
			th[i].trigger := 0.0;
			th[i].dataptr := th[i].data;
			th[i].tracknum := i;
		 end;

		// Record starting time
    curtime := timeGetTime;
		ms.starttime := curtime;
    starttime := curtime;

		// Loop until we've been asked to stop
		first_pass := True;
		while (not ms.stop_requested) do
		 begin
			// See if we've been asked to seek and we're not already doing it
			if ms.seeking and not seeking then
			 begin
  			seeking := True;
				// Turn all notes off
        all_notes_off;
				goto BeginPlayback;		// Aaaah!  A goto!!
			 end;
			// Get current time
			if not seeking then curtime := timeGetTime;
			// Schedule the next "wakeup time" MAX_MIDI_WAIT milliseconds into the future
			nexttrigger := curtime + MAX_MIDI_WAIT;
			// See if any tracks have pending data
			tracks_active := False;
			for i := 0 to tracks - 1 do
			 begin
				// Continue on to the next track if this track is disabled
				if (th[i].enabled) then
					tracks_active := True
				else
					continue;

				// Disable this track if its data pointer has reached or exceeded the track length
				if longword(th[i].dataptr) >= longword(th[i].data) + th[i].length then
				 begin
					th[i].enabled := FALSE;
					continue;
				 end;

				// Read the first event's delta time
				if first_pass then
				 begin
					th[i].dt := read_vlq_mem(th[i]) * ms.tick_length;
					th[i].trigger := curtime + th[i].dt;
				 end;

				// Process MIDI events until one is scheduled for a time in the future
				while curtime >= th[i].trigger do
				 begin
					// Process the event for this track
					ms.curtime := curtime;
					if process_midi_event(th[i]) then
					 begin
						ms.stop_requested := True;
						break;
					 end;

					// Check for end of track
					if not (th[i].enabled) then	break;

					// Read the next event's delta time
					th[i].dt := read_vlq_mem(th[i]) * ms.tick_length;
					th[i].trigger := th[i].trigger + th[i].dt;
				 end;
				// Check for end of track
				if not th[i].enabled then continue;

				// See if this track's trigger is the more recent than the next scheduled trigger
				if th[i].trigger < nexttrigger then
         nexttrigger := th[i].trigger;
			 end;
			first_pass := False;
			if not seeking then	curtime := timeGetTime;
			// Check to see if a pause has been requested
			if ms.paused then
			 begin
				// Turn off the notes being played
				all_notes_off;
				// Save the time we paused at
				pausetime := curtime;
				// Wait until we're unpaused or not playing anymore
				while True do
				 begin
					Sleep(100);
					if not (ms.paused) or not (ms.playing) or ms.stop_requested then
						break;
				 end;
				// If we've been asked to stop.. break from the main playback loop
				if not (ms.playing) or ms.stop_requested then
					break;
				// We need to resume.  Correct the start and trigger times to be up-to-date!
				curtime := timeGetTime;
				starttime := starttime + curtime - pausetime;
				nexttrigger := nexttrigger + curtime - pausetime;
				for i := 0 to tracks - 1 do
					th[i].trigger := th[i].trigger + curtime - pausetime;
			 end;

			if not seeking then
       ms.elapsedtime := trunc((curtime - starttime) / 1000.0);

			// Wait until the next trigger time
			if seeking then
			 begin
				curtime := nexttrigger;
				// See if the seek is done
				if curtime - starttime >= ms.seek_to then
				 begin
					// Seek is done - correct timing values
					tmptime := curtime;
					curtime := timeGetTime;
					timediff := tmptime - curtime;
					for i := 0 to mh.num_tracks - 1 do
						th[i].trigger := th[i].trigger - timediff;
					// Calculate corrected start time and elapsed time
          starttime := starttime - timediff;
					ms.starttime := starttime;
					ms.elapsedtime := trunc((curtime - starttime) / 1000.0);
					ms.curtime := curtime;
					// Reset seek state flags
          ms.analyzing := False;
          ms.seeking := False;
					seeking := False;

					// Set channel values
					for i := 0 to 15 do
					 begin
						if (ms.channels[i].changed_pitch_bend >= 0) then
							set_channel_pitch(i, ms.channels[i].changed_pitch_bend);
						// Set program changes
						if (ms.channels[i].normal_program >= 0) then
							set_channel_program(i, ms.channels[i].normal_program);
					 end;
				 end;
			 end
			else
			 begin
				while (curtime < nexttrigger) and not ms.stop_requested do
				 begin
					Sleep(trunc(nexttrigger - curtime));
					curtime := timeGetTime;
         end;
				if not tracks_active then
				 begin
					ms.finished_naturally := TRUE;
					break;
				 end;
			 end;
		 end;
		// Break out of looping if a stop was requested
		if ms.stop_requested then
			break;
		// If the looping checkbox is checked, loop another time
		if MIDILoop then
			inc(ms.loop_count);
	 end;
	// PLAYBACK HAS STOPPED
  //Restore default MMTimers precision
  timeEndPeriod(PeriodMin);
	// Turn off all the notes, reset the MIDI-out device, and close it
	all_notes_off;
	Sleep(50);
	midiOutReset(hout);
	close_midi_out;

	// Reset the track info displays
	for i := 0 to 15 do
	 begin
    FillChar(ms.channels[i].notes,sizeof(ms.channels[i].notes),0);
		ms.channels[i].note_count := 0;
	 end;


	ms.playing := False;
	PostMessage(hwndApp, WM_USER, 0, 0);
end;

function MIDIThreadActive:boolean;
var
 ExCode:DWORD;
begin
Result := (MIDIThreadH <> 0) and
          GetExitCodeThread(MIDIThreadH,ExCode) and
          (ExCode = STILL_ACTIVE);
if not Result then
 if MIDIThreadH <> 0 then
  begin
   CloseHandle(MIDIThreadH);
   MIDIThreadH := 0
  end
end;

procedure WaitForMIDIThreadExit;
var
 ExCode:DWORD;
begin
if MIDIThreadH = 0 then exit;
repeat
 if not GetExitCodeThread(MIDIThreadH,ExCode) then break;
 if ExCode = STILL_ACTIVE then Sleep(10)
until ExCode <> STILL_ACTIVE;
CloseHandle(MIDIThreadH);
MIDIThreadH := 0
end;

procedure StopMIDIThread;
var
 msg:TMsg;
 i:integer;
begin
if MIDIThreadActive then
 begin
  ms.stop_requested := True; //,  event
  WaitForMIDIThreadExit;
  while not PeekMessage(msg,hwndApp,
                      WM_USER,WM_USER,PM_REMOVE) do Sleep(10);
 end;
	// Dump any previously-loaded MIDI text events
	kill_all_midi_text;

	// Unload any previously-loaded MIDI data
	if mh.num_tracks <> 0 then
		for i := 0 to mh.num_tracks - 1 do
		 begin
			th[i].dataptr := nil;
			if th[i].data <> nil then FreeMem(th[i].data);
      FillChar(th[i],sizeof(track_header_t),0);
		 end;
  mh.num_tracks := 0   
end;

end.
